#  This file is managed by 'repo_helper'. Don't edit it directly.

__all__ = ["extras_require"]

extras_require = {
		"limiter": ["cachecontrol[filecache]>=0.12.6", "lockfile>=0.12.2"],
		"all": ["cachecontrol[filecache]>=0.12.6", "lockfile>=0.12.2"]
		}
